import org.jfree.chart.ChartFactory
import org.jfree.chart.ChartUtilities
import org.jfree.chart.plot.DefaultDrawingSupplier
import org.jfree.chart.plot.PlotOrientation
import org.jfree.chart.renderer.category.BarRenderer
import org.jfree.chart.renderer.category.StandardBarPainter
import org.jfree.data.category.CategoryDataset
import org.jfree.data.general.PieDataset
import org.jfree.data.time.TimeSeriesCollection
import java.awt.BasicStroke
import java.awt.Color
import java.awt.Paint
import java.awt.geom.Ellipse2D


chart = new ChartTool()

/**
 * Velocity tool for chart drawing with the JIRA PDF View Plugin.
 * <p>
 * Each drawing method receives two arguments:
 * <ol>
 * <li>dataset: its type depends on the actual chart to be drawn</li>
 * <li>customizer: this is a callback with a single argument (the chart object itself) which can make customization
 *     after the chart was created. Customization includes changing rendering parameters like color or fonts, changing labels
 *     adding markers and so on.<li>
 * <p>
 * All chart drawing menthods return chart images as BASE64-encoded strings (PNG format). Those can be inserted into
 * the PDF templates (<tt>issue-fo.vm</tt>) by defining images with data URIs:
 * <code>
 * 	&lt;fo:external-graphic content-height="4cm" src="data:image/png;base64,$chart.drawPieChart(dataset, customizer, 400, 300)"/&gt;
 * </code>
 * Suggestion: look at the concrete charts as examples. Sample code will make it clear.
 */
public class ChartTool {
	public String drawTimeSeriesChart(TimeSeriesCollection dataset, customizer, width, height) {
		// create chart
		def chart = ChartFactory.createTimeSeriesChart(null, null, null, dataset, false, false, false)
		chart.plot.backgroundPaint = RenderingUtils.BACKGROUND_COLOR

		customizer(chart)

		// render chart
		ChartUtilities.encodeAsPNG(chart.createBufferedImage(width, height)).encodeBase64().toString()
	}

	public String drawPieChart(PieDataset dataset, customizer, width, height) {
		// create chart
		def chart = ChartFactory.createPieChart(null, dataset, false, false, false)
		chart.plot.backgroundPaint = RenderingUtils.BACKGROUND_COLOR
		chart.plot.shadowPaint = null
		chart.plot.labelOutlinePaint = null
		chart.plot.labelShadowPaint = RenderingUtils.BACKGROUND_COLOR
		chart.plot.labelBackgroundPaint = RenderingUtils.BACKGROUND_COLOR

		customizer(chart)

		// render chart
		ChartUtilities.encodeAsPNG(chart.createBufferedImage(width, height)).encodeBase64().toString()
	}

	public String drawBarChart(CategoryDataset dataset, customizer, width, height) {
		// create chart
		def chart = ChartFactory.createBarChart(null, null, null, dataset, PlotOrientation.VERTICAL, false, false, false)
		chart.plot.backgroundPaint = RenderingUtils.BACKGROUND_COLOR

		chart.plot.renderer.barPainter = new StandardBarPainter()
		chart.plot.renderer.drawBarOutline = true
		chart.plot.renderer.shadowVisible = false

		customizer(chart)

		// render chart
		ChartUtilities.encodeAsPNG(chart.createBufferedImage(width, height)).encodeBase64().toString()
	}
}

/**
 * Drawing supplier to provide a color gradient to paint items of unknown number.
 */
public class GradientDrawingSupplier extends DefaultDrawingSupplier {
	private float max
	private int i

	GradientDrawingSupplier(max) {
		this.max = max
	}

	Paint getNextPaint() {
		return RenderingUtils.getGradientPaint(++i/max)
	}
}

/**
 * Drawing supplier to provide a color gradient to paint bars of unknown number.
 */
public class GradientBarRenderer extends BarRenderer {
	private float max

	GradientBarRenderer(max) {
		this.max = max

		shadowVisible = false
		barPainter = new StandardBarPainter()
	}

	Paint getItemPaint(int row, int col) {
		return RenderingUtils.getGradientPaint(col/max)
	}
}

private class RenderingUtils {
	static def BACKGROUND_COLOR = new Color(0xffffff)
	static def MARKER_COLOR = new Color(0xaaaaaa)

	static def LINE_STROKE = new BasicStroke(3.0f, 1, 1)
	static def MARKER_STROKE = new BasicStroke(1.0f, BasicStroke.CAP_BUTT, BasicStroke.JOIN_MITER, 10.0f, [ 10.0f ] as float[], 0.0f)

	static def CIRCLE_SHAPE = new Ellipse2D.Float(-2.0f, -2.0f, 4.0f, 4.0f)

	/**
	 * Gets a paint based on value of t (in [0 ... 1]) that can be used to paint items if their count is not fixed.
	 */
	static Paint getGradientPaint(t) {
		return new Color(1.0f, (float)(0.6 + 0.32*t), (float)(0.8*t))
	}
}

