import org.apache.log4j.Logger
import org.apache.commons.lang.StringUtils
import org.apache.commons.lang.math.NumberUtils
import com.google.common.collect.Table
import com.google.common.collect.HashBasedTable
import com.atlassian.jira.component.ComponentAccessor
import com.atlassian.jira.issue.Issue
import com.atlassian.jira.issue.link.IssueLinkManager
import com.atlassian.jira.issue.search.SearchProvider
import com.atlassian.jira.issue.search.SearchRequest
import com.atlassian.jira.issue.search.SearchRequestManager
import com.atlassian.jira.web.bean.PagerFilter

if (!binding.variables.containsKey("searchRequest")) {
	searchRequest = null // set it explicitly to avoid "not a property" exceptions
}
def issueCollector = new IssueCollector(user, searchRequest)
leftIssues = issues
topIssues = issueCollector.getTopIssues()
if(topIssues == null) {
	// use the passed issues both in the top row and the left column
	topIssues = issues
}

traceabilityMatrix = new TraceabilityMatrixTool(user, leftIssues, topIssues, linkTypeNames)

/**
 * Helper class to collect the issues for the top row.
 */
private class IssueCollector {
	def log = Logger.getLogger(this.getClass())

	private user
	private searchRequest

	public IssueCollector(user, searchRequest) {
		this.user = user
		this.searchRequest = searchRequest
	}

	/**
	 * Parses the filter ID from the description of the used search request.
	 * The ID is expected after the last '#' and before the next ' ' character.
	 */
	private int getTopSearchRequestId() {
		def idAsString = StringUtils.trimToNull(StringUtils.substringBefore(StringUtils.substringAfterLast(searchRequest.description, "#"), " "))
		return NumberUtils.toInt(idAsString)
	}

	public List<Issue> getTopIssues() {
		log.debug("Searching for top issues")

		if(searchRequest == null) {
			log.debug("No left filter available")
			return null;
		}

		def searchRequestId = getTopSearchRequestId()
		if(searchRequestId == 0) {
			log.debug("No top filter defined")
			return null
		}
		log.debug("Using top filter #" + searchRequestId)

		def searchRequest = ComponentAccessor.getComponentOfType(SearchRequestManager.class).getSearchRequestById(user, searchRequestId)
		if(searchRequest == null) {
			log.warn("Filter #" + searchRequestId + " not found")
			return null
		}
		log.debug("Filter found: <" + searchRequest.name + ">")

		def searchResults = ComponentAccessor.getComponentOfType(SearchProvider.class).search(searchRequest.query, user, PagerFilter.getUnlimitedFilter())
		if(searchResults == null) {
			log.warn("No results found")
			return null
		}
		log.debug("Results found: " + searchResults.total)

		return searchResults.issues
	}
}

/**
 * Velocity tool to help rendering traceability matrices with the JIRA PDF View Plugin.
 * @see traceability-matrix-fo.vm as the actual template.
 */
public class TraceabilityMatrixTool {
	def log = Logger.getLogger(this.getClass())

	def issueLinkManager = ComponentAccessor.getIssueLinkManager()

	private List<Issue> leftIssues
	private List<Issue> topIssues
	private Table<String, String, Dependency> dependencies = HashBasedTable.create()

	public TraceabilityMatrixTool(user, leftIssues, topIssues, linkTypeNames) {
		log.debug("Computing dependencies for " + leftIssues.size() + " vs. " + topIssues.size() + " issues")

		this.leftIssues = leftIssues
		this.topIssues = topIssues

		// collect dependencies
		collectDependencies(user, leftIssues, linkTypeNames)
		if(topIssues != leftIssues) {
			collectDependencies(user, topIssues, linkTypeNames)
		}
	}

	private def collectDependencies(user, issues, linkTypeNames) {
		for(issue in issues) {
			log.debug("Collecting dependencies for <" + issue.key + ">")
			def linkCollection = issueLinkManager.getLinkCollection(issue, user)
			for(linkTypeName in linkTypeNames.tokenize(',')) {
				linkTypeName = StringUtils.trim(linkTypeName)
				addDependencies(issue, linkCollection, linkTypeName)
			}
		}
	}

	private def addDependencies(issue, linkCollection, linkTypeName) {
		def outwardIssues = linkCollection.getOutwardIssues(linkTypeName)

		if(outwardIssues) {
			for(outwardIssue in outwardIssues) {
				addDependency(issue.key, outwardIssue.key, 'O')
				addDependency(outwardIssue.key, issue.key, 'I')
			}
		}
	}

	private def addDependency(String fromKey, String toKey, String direction) {
		def dependency = dependencies.get(fromKey, toKey)
		if(dependency) {
			// make it bi-directional if there is already a dependency found
			dependency = new Dependency(direction: 'B')
		} else {
			dependency = new Dependency(direction: direction)
		}

		dependencies.put(fromKey, toKey, dependency)
	}

	public List<Issue> getLeftIssues() {
		return leftIssues
	}

	public List<Issue> getTopIssues() {
		return topIssues
	}

	public int getDependencyCount(String key) {
		return dependencies.column(key).size();
	}

	public Dependency getDependency(String fromKey, String toKey) {
		return dependencies.get(fromKey, toKey)
	}

	public static class Dependency {
		/** 'O' for outward, 'I' for inward or 'B' for bi-directional. */
		def direction

		public String toString() {
			return direction
		}
	}
}
